Attribute VB_Name = "modDrawToolTip"
Option Explicit

Private Declare Function GetWindowRect Lib "user32" (ByVal hwnd As Long, lpRect As RECT) As Long
Private Declare Function GetCursorPos Lib "user32" (lpPoint As POINTAPI) As Long
Private Declare Function MoveWindow Lib "user32" (ByVal hwnd As Long, ByVal x As Long, ByVal y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal bRepaint As Long) As Long
Private Declare Function ShowWindow Lib "user32" (ByVal hwnd As Long, ByVal nCmdShow As Long) As Long
Private Declare Function BeginPaint Lib "user32" (ByVal hwnd As Long, lpPaint As PAINTSTRUCT) As Long
Private Declare Function SetBkMode Lib "gdi32" (ByVal hdc As Long, ByVal nBkMode As Long) As Long
Private Declare Function SelectObject Lib "gdi32" (ByVal hdc As Long, ByVal hObject As Long) As Long
Private Declare Function DeleteObject Lib "gdi32" (ByVal hObject As Long) As Long
Private Declare Function DrawText Lib "user32" Alias "DrawTextA" (ByVal hdc As Long, ByVal lpStr As String, ByVal nCount As Long, lpRect As RECT, ByVal wFormat As Long) As Long
Private Declare Function EndPaint Lib "user32" (ByVal hwnd As Long, lpPaint As PAINTSTRUCT) As Long
Private Declare Function GetClientRect Lib "user32" (ByVal hwnd As Long, lpRect As RECT) As Long
Private Declare Function CreateRectRgn Lib "gdi32" (ByVal X1 As Long, ByVal Y1 As Long, ByVal X2 As Long, ByVal Y2 As Long) As Long
Private Declare Function CreateRoundRectRgn Lib "gdi32" (ByVal X1 As Long, ByVal Y1 As Long, ByVal X2 As Long, ByVal Y2 As Long, ByVal X3 As Long, ByVal Y3 As Long) As Long
Private Declare Function FrameRgn Lib "gdi32" (ByVal hdc As Long, ByVal hRgn As Long, ByVal hBrush As Long, ByVal nWidth As Long, ByVal nHeight As Long) As Long
Private Declare Function CreateBrushIndirect Lib "gdi32" (lpLogBrush As LOGBRUSH) As Long
Private Declare Function FillRgn Lib "gdi32" (ByVal hdc As Long, ByVal hRgn As Long, ByVal hBrush As Long) As Long
Private Declare Function SetTextColor Lib "gdi32" (ByVal hdc As Long, ByVal crColor As Long) As Long
Private Declare Function GetStockObject Lib "gdi32" (ByVal nIndex As Long) As Long
Private Declare Function GetSysColor Lib "user32" (ByVal nIndex As Long) As Long
Private Declare Function CreateFontIndirect Lib "gdi32" Alias "CreateFontIndirectA" (lpLogFont As LOGFONT) As Long

Private Const COLOR_INFOTEXT = 23
Private Const COLOR_INFOBK = 24
Private Const COLOR_GRAYTEXT = 17
Private Const COLOR_3DLIGHT = 22

Private Const SW_SHOWNA         As Long = 8
Private Const TRANSPARENT       As Long = 1

Private Const DT_SINGLELINE = &H20
Private Const DT_NOCLIP = &H100
Private Const DT_CENTER = &H1
Private Const DT_VCENTER = &H4
Private Const DT_WORDBREAK = &H10
Private Const DT_CALCRECT = &H400

Private Const BS_SOLID = 0
Private Const HS_CROSS = 4

Private Const DKGRAY_BRUSH = 3

Private Const LF_FACESIZE As Long = 32

Private Type LOGBRUSH
    lbStyle As Long
    lbColor As Long
    lbHatch As Long
End Type

Private Type RECT
    Left As Long
    Top As Long
    Right As Long
    Bottom As Long
End Type

Private Type PAINTSTRUCT
    hdc As Long
    fErase As Long
    rcPaint As RECT
    fRestore As Long
    fIncUpdate As Long
    rgbReserved(32) As Byte
End Type

Private Type LOGFONT
    lfHeight As Long
    lfWidth As Long
    lfEscapement As Long
    lfOrientation As Long
    lfWeight As Long
    lfItalic As Byte
    lfUnderline As Byte
    lfStrikeOut As Byte
    lfCharSet As Byte
    lfOutPrecision As Byte
    lfClipPrecision As Byte
    lfQuality As Byte
    lfPitchAndFamily As Byte
    lfFaceName(1 To LF_FACESIZE) As Byte
End Type

Private Type POINTAPI
    x As Long
    y As Long
End Type

Private Const CURSOR_OFFSET_X As Long = 10
Private Const CURSOR_OFFSET_Y As Long = 15

Private Const FONT_NAME = "Tahoma"
Private Const FONT_SIZE = 13

Private Const TEXT_OFFSET_X As Long = 5
Private Const TEXT_OFFSET_Y As Long = 3

Public Sub DrawToolTip(ByRef pobjTip As Tooltip)
Dim udtPS       As PAINTSTRUCT
    udtPS = CreateDrawingArea(pobjTip)
    DrawBalloon pobjTip, udtPS.hdc
    EndPaint pobjTip.hwnd, udtPS
End Sub

Public Function CreateDrawingArea(ByRef pobjTip As Tooltip) As PAINTSTRUCT
Dim udtPS           As PAINTSTRUCT
Dim udtTipRC        As RECT
Dim udtCurPosPT     As POINTAPI
Dim lngFont         As Long
Dim lngFontOld      As Long

    'Position Window To Cursor
    GetWindowRect pobjTip.hwnd, udtTipRC
    GetCursorPos udtCurPosPT
    With udtTipRC
        .Left = udtCurPosPT.x + CURSOR_OFFSET_X
        .Top = udtCurPosPT.y + CURSOR_OFFSET_Y
        MoveWindow pobjTip.hwnd, .Left, .Top, .Right - .Left, .Bottom - .Top, False
    End With
    
    'Prepare The DC For Drawing
    BeginPaint pobjTip.hwnd, udtPS
    CreateDrawingArea = udtPS
    
    'Set Background To Be Transparent
    SetBkMode udtPS.hdc, TRANSPARENT
    
    'Create font
    lngFont = CreateFont(FONT_NAME, FONT_SIZE)
    lngFontOld = SelectObject(udtPS.hdc, lngFont)
  
    'Draw Text Window
    DrawText udtPS.hdc, pobjTip.Text, Len(pobjTip.Text), udtTipRC, DT_VCENTER + DT_NOCLIP + DT_CALCRECT
    
    'Resize Text Window To Fit Balloon
    With udtTipRC
        .Right = .Right + 2 * TEXT_OFFSET_X
        .Bottom = .Bottom + 2 * TEXT_OFFSET_Y
        ShowWindow pobjTip.hwnd, SW_SHOWNA
        MoveWindow pobjTip.hwnd, .Left, .Top, .Right - .Left, .Bottom - .Top, True
    End With
    
    'Restore the Old Font
    SelectObject udtPS.hdc, lngFontOld
    DeleteObject lngFont
End Function

Private Sub DrawBalloon(ByRef pobjTip As Tooltip, ByVal plnghDC As Long)
Dim lngFont         As Long
Dim lngOldFont      As Long
Dim udtTipRC        As RECT
Dim udtBalloonRC    As RECT
Dim udtTextRC       As RECT

Dim lngRgnBalloon   As Long

    'Get Window RECT
    GetClientRect pobjTip.hwnd, udtTipRC
    
    'Setup Balloon RECT
    LSet udtBalloonRC = udtTipRC
    
    'Setup Text RECT
    LSet udtTextRC = udtBalloonRC
    With udtTextRC
        .Left = .Left + TEXT_OFFSET_X
        .Right = .Right - TEXT_OFFSET_X
        .Top = .Top + TEXT_OFFSET_Y
        .Bottom = .Bottom - TEXT_OFFSET_Y
    End With
    
    'Since tip background is transparent, we don't know if x and y coords are correct
    'fill tip background color, testing purposes only to
    'check x and y coords of other stuff drawn

    'FillRegion plnghDC, CreateRegion(udtTipRC, 0), RGB(0, 255, 0)
    
    'Create Balloon Region
    lngRgnBalloon = CreateRegion(udtBalloonRC, 0)
    FillRegion plnghDC, lngRgnBalloon, GetSysColor(COLOR_INFOBK)
    
    'Add Border To Balloon
    FrameRgn plnghDC, lngRgnBalloon, GetStockObject(DKGRAY_BRUSH), 1, 1
    
    'Create font
    lngFont = CreateFont(FONT_NAME, FONT_SIZE)
    lngOldFont = SelectObject(plnghDC, lngFont)
    
    'Add Text
    DrawCustomText plnghDC, pobjTip.Text, GetSysColor(COLOR_INFOTEXT), udtTextRC
    
    'Restore the Old Font
    SelectObject plnghDC, lngOldFont
    DeleteObject lngFont
End Sub

Private Sub FillRegion(ByVal plnghDC As Long, ByVal plngRegion As Long, ByVal plngColor As Long)
Dim lngBrush    As Long
Dim typLB       As LOGBRUSH
    With typLB
        .lbColor = plngColor
        .lbStyle = BS_SOLID
        .lbHatch = HS_CROSS
    End With
    lngBrush = CreateBrushIndirect(typLB)
    FillRgn plnghDC, plngRegion, lngBrush
    DeleteObject lngBrush
End Sub

Private Sub DrawCustomText(ByVal plnghDC As Long, ByVal pstrText As String, ByVal plngColor As Long, ByRef pudtRC As RECT)
    SetTextColor plnghDC, plngColor
    DrawText plnghDC, pstrText, Len(pstrText), pudtRC, DT_VCENTER + DT_NOCLIP
End Sub

Private Function CreateRegion(ByRef pudtRC As RECT, Optional ByVal plngCurveSize As Long) As Long
Dim lngRegion   As Long
    With pudtRC
        If plngCurveSize > 0 Then
            lngRegion = CreateRoundRectRgn(.Left, .Top, .Right, .Bottom, plngCurveSize, plngCurveSize)
        Else
            lngRegion = CreateRectRgn(.Left, .Top, .Right, .Bottom)
        End If
    End With
    CreateRegion = lngRegion
End Function

Private Function CreateFont(ByVal pstrName As String, ByVal plngSize As Long) As Long
Dim lngIndex    As Long
Dim udtLF       As LOGFONT
    With udtLF
        .lfHeight = plngSize
        For lngIndex = 1 To Len(pstrName)
            .lfFaceName(lngIndex) = Asc(Mid(pstrName, lngIndex, 1))
        Next lngIndex
    End With
    CreateFont = CreateFontIndirect(udtLF)
End Function
